from typing import Generic, Protocol, TypeVar

from lonelypsp.tracing.stateless.notify import (
    StatelessTracingNotifyOnReceived,
    StatelessTracingNotifyStart,
)

InitializerTcontra = TypeVar("InitializerTcontra", contravariant=True)


class StatelessTracingSubscriberRoot(Generic[InitializerTcontra], Protocol):
    """Contains a way to produce the start of each stateless operation
    from the subscribers perspective. Many of these protocols may have
    overlapping names, so this cannot just absorb all the protocols
    """

    def notify(self, initializer: InitializerTcontra, /) -> StatelessTracingNotifyStart:
        """About to start the process of sending a NOTIFY message to the broadcaster
        and receiving a RESPOND_NOTIFY
        """


class StatelessTracingBroadcasterRoot(Generic[InitializerTcontra], Protocol):
    """Contains a way to produce the start of each stateless operation
    from the broadcasters perspective.  Many of these protocols may have
    overlapping names, so this cannot just absorb all the protocols
    """

    def receive_notify(
        self, initializer: InitializerTcontra, /
    ) -> StatelessTracingNotifyOnReceived:
        """About to start the process of receiving a NOTIFY message from the subscriber
        and sending a RESPOND_NOTIFY
        """



from typing import TYPE_CHECKING, Protocol, Type, Union


class SyncReadableBytesIOA(Protocol):
    """A type that represents a stream that can be read synchronously"""

    def read(self, n: int) -> bytes:
        """Reads n bytes from the file-like object"""
        raise NotImplementedError()


class SyncReadableBytesIOB(Protocol):
    """A type that represents a stream that can be read synchronously"""

    def read(self, n: int, /) -> bytes:
        """Reads n bytes from the file-like object"""
        raise NotImplementedError()


SyncReadableBytesIO = Union[SyncReadableBytesIOA, SyncReadableBytesIOB]


class SyncWritableBytesIO(Protocol):
    """A type that represents a stream that can be written synchronously"""

    def write(self, b: Union[bytes, bytearray], /) -> int:
        """Writes the given bytes to the file-like object"""
        raise NotImplementedError()


class Closeable(Protocol):
    """Represents something that can be closed"""

    def close(self) -> None:
        """Closes the file-like object"""


class SyncTellableBytesIO(Protocol):
    """A type that represents a stream with a synchronous tell method"""

    def tell(self) -> int:
        """Returns the current position in the file"""
        raise NotImplementedError()


class SyncSeekableBytesIO(Protocol):
    """A type that represents a stream with a synchronous seek method"""

    def seek(self, offset: int, whence: int = 0) -> int:
        """Seeks to a position in the file"""
        raise NotImplementedError()


class SyncStandardIOA(
    SyncReadableBytesIOA, SyncTellableBytesIO, SyncSeekableBytesIO, Protocol
): ...


class SyncStandardIOB(
    SyncReadableBytesIOA, SyncTellableBytesIO, SyncSeekableBytesIO, Protocol
): ...


SyncStandardIO = Union[SyncStandardIOA, SyncStandardIOB]


class PreallocatedBytesIO:
    """Convenience object that acts like BytesIO, but you can specify the length
    in advance
    """

    def __init__(self, length: int) -> None:
        self.buffer = bytearray(length)
        self.pos = 0

    def read(self, n: int) -> bytes:
        if n < 0:
            n = len(self.buffer) - self.pos

        result = self.buffer[self.pos : self.pos + n]
        self.pos = min(self.pos + n, len(self.buffer))
        return result

    def write(self, b: Union[bytes, bytearray], /) -> int:
        end_index = self.pos + len(b)
        if end_index > len(self.buffer):
            raise ValueError("writing past the end of the buffer")

        self.buffer[self.pos : end_index] = b
        self.pos = end_index

        return len(b)

    def tell(self) -> int:
        return self.pos

    def seek(self, offset: int, whence: int = 0) -> int:
        if whence == 0:
            self.pos = offset
        elif whence == 1:
            self.pos += offset
        elif whence == 2:
            self.pos = len(self.buffer) + offset
        else:
            raise ValueError("invalid whence value")

        if self.pos < 0:
            self.pos = 0
        elif self.pos > len(self.buffer):
            self.pos = len(self.buffer)
        return self.pos

    def close(self) -> None:
        pass


if TYPE_CHECKING:
    _: Type[SyncStandardIO] = PreallocatedBytesIO



from typing import TYPE_CHECKING, Collection, List, Literal, Optional, Type, Union

from lonelypsp.compat import fast_dataclass
from lonelypsp.stateful.constants import (
    BroadcasterToSubscriberStatefulMessageType,
    PubSubStatefulMessageFlags,
)
from lonelypsp.stateful.generic_parser import B2S_MessageParser
from lonelypsp.stateful.parser_helpers import parse_simple_headers
from lonelypsp.stateful.serializer_helpers import (
    MessageSerializer,
    serialize_simple_message,
)
from lonelypsp.sync_io import SyncReadableBytesIO


@fast_dataclass
class B2S_Missed:
    """
    B2S = Broadcaster to Subscriber
    See the type enum documentation for more information on the fields
    """

    type: Literal[BroadcasterToSubscriberStatefulMessageType.MISSED]
    """discriminator value"""

    authorization: Optional[str]
    """recovery: websocket:<nonce>:<ctr>
    
    an empty string is reinterpreted as None for consistency between
    minimal headers mode and expanded headers mode
    """

    tracing: bytes
    """the tracing data, which may be empty"""

    topic: bytes
    """the topic the possibly missed message was sent to"""


_headers: Collection[str] = (
    "authorization",
    "x-tracing",
    "x-topic",
)


class B2S_MissedParser:
    """Satisfies B2S_MessageParser[B2S_Missed]"""

    @classmethod
    def relevant_types(cls) -> List[BroadcasterToSubscriberStatefulMessageType]:
        return [BroadcasterToSubscriberStatefulMessageType.MISSED]

    @classmethod
    def parse(
        cls,
        flags: PubSubStatefulMessageFlags,
        type: BroadcasterToSubscriberStatefulMessageType,
        payload: SyncReadableBytesIO,
    ) -> B2S_Missed:
        assert type == BroadcasterToSubscriberStatefulMessageType.MISSED

        headers = parse_simple_headers(flags, payload, _headers)

        authorization_bytes = headers["authorization"]
        authorization: Optional[str] = None
        if authorization_bytes != b"":
            try:
                authorization = authorization_bytes.decode("utf-8")
            except UnicodeDecodeError:
                raise ValueError("authorization must be a utf-8 string")

        tracing = headers["x-tracing"]
        topic = headers["x-topic"]

        return B2S_Missed(
            type=type,
            authorization=authorization,
            tracing=tracing,
            topic=topic,
        )


if TYPE_CHECKING:
    _: Type[B2S_MessageParser[B2S_Missed]] = B2S_MissedParser


def serialize_b2s_missed(
    msg: B2S_Missed, /, *, minimal_headers: bool
) -> Union[bytes, bytearray]:
    """Satisfies MessageSerializer[B2S_Missed]"""
    return serialize_simple_message(
        type=msg.type,
        header_names=_headers,
        header_values=(
            msg.authorization.encode("utf-8") if msg.authorization is not None else b"",
            msg.tracing,
            msg.topic,
        ),
        payload=b"",
        minimal_headers=minimal_headers,
    )


if TYPE_CHECKING:
    __: MessageSerializer[B2S_Missed] = serialize_b2s_missed



# ruff: noqa: T201
"""Verifies that with cancel_and_check it is possible to use get() and put(),
though it is rather perplexing
"""

import asyncio

from lonelypsp.util.cancel_and_check import cancel_and_check
from lonelypsp.util.drainable_asyncio_queue import DrainableAsyncioQueue


async def _test() -> None:
    await _test_get_no_yield()
    print(".", end="", flush=True)
    await _test_get_with_1_yield()
    print(".", end="", flush=True)
    await _test_get_with_2_yield()
    print(".", end="", flush=True)
    await _test_put_no_yield()
    print(".", end="", flush=True)
    await _test_put_with_1_yield()
    print(".", end="", flush=True)
    await _test_put_with_2_yield()
    print(".", end="", flush=True)
    print("done")


async def _test_get_no_yield() -> None:
    q: DrainableAsyncioQueue[int] = DrainableAsyncioQueue()
    getter = asyncio.create_task(q.get())
    canceler = asyncio.create_task(cancel_and_check(getter))
    q.put_nowait(1)
    assert (await canceler) == 1
    assert q.drain() == []


async def _test_get_with_1_yield() -> None:
    q: DrainableAsyncioQueue[int] = DrainableAsyncioQueue()
    getter = asyncio.create_task(q.get())
    canceler = asyncio.create_task(cancel_and_check(getter))
    await asyncio.sleep(0)
    q.put_nowait(1)
    assert (await canceler) == 1
    assert q.drain() == []


async def _test_get_with_2_yield() -> None:
    q: DrainableAsyncioQueue[int] = DrainableAsyncioQueue()
    getter = asyncio.create_task(q.get())
    canceler = asyncio.create_task(cancel_and_check(getter))
    await asyncio.sleep(0)
    await asyncio.sleep(0)
    q.put_nowait(1)
    result = await canceler
    assert result is None, f"{result=}"
    assert q.drain() == [1]


async def _test_put_no_yield() -> None:
    q: DrainableAsyncioQueue[int] = DrainableAsyncioQueue(max_size=1)
    q.put_nowait(1)
    putter = asyncio.create_task(q.put(2))
    canceler = asyncio.create_task(cancel_and_check(putter, False))
    assert q.get_nowait() == 1
    assert (await canceler) is None  # None is the result of put() on success
    assert q.drain() == [2]


async def _test_put_with_1_yield() -> None:
    q: DrainableAsyncioQueue[int] = DrainableAsyncioQueue(max_size=1)
    q.put_nowait(1)
    putter = asyncio.create_task(q.put(2))
    canceler = asyncio.create_task(cancel_and_check(putter, False))
    await asyncio.sleep(0)
    assert q.get_nowait() == 1
    assert (await canceler) is None  # None is the result of put() on success
    assert q.drain() == [2]


async def _test_put_with_2_yield() -> None:
    q: DrainableAsyncioQueue[int] = DrainableAsyncioQueue(max_size=1)
    q.put_nowait(1)
    putter = asyncio.create_task(q.put(2))
    canceler = asyncio.create_task(cancel_and_check(putter, False))
    await asyncio.sleep(0)
    await asyncio.sleep(0)
    assert q.get_nowait() == 1
    assert (await canceler) is False  # default value, meaning put() failed
    assert q.drain() == []


if __name__ == "__main__":
    asyncio.run(_test())



from typing import TYPE_CHECKING, Optional, Type

from lonelypsp.auth.config import (
    AuthResult,
    ToBroadcasterAuthConfig,
    ToSubscriberAuthConfig,
)
from lonelypsp.auth.set_subscriptions_info import SetSubscriptionsInfo
from lonelypsp.stateful.messages.configure import S2B_Configure
from lonelypsp.stateful.messages.confirm_configure import B2S_ConfirmConfigure
from lonelypsp.stateful.messages.continue_receive import S2B_ContinueReceive
from lonelypsp.stateful.messages.disable_zstd_custom import B2S_DisableZstdCustom
from lonelypsp.stateful.messages.enable_zstd_custom import B2S_EnableZstdCustom
from lonelypsp.stateful.messages.enable_zstd_preset import B2S_EnableZstdPreset
from lonelypsp.stateless.make_strong_etag import StrongEtag


class ToBroadcasterNoneAuth:
    """Sets up and allows a broadcaster that does not block any incoming requests.

    In order for this to be secure it must only be possible for trusted clients
    to connect to the server (e.g., by setting up TLS mutual auth at the binding
    level)
    """

    async def setup_to_broadcaster_auth(self) -> None: ...
    async def teardown_to_broadcaster_auth(self) -> None: ...

    async def authorize_subscribe_exact(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
    ) -> Optional[str]:
        return None

    async def is_subscribe_exact_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return AuthResult.OK

    async def authorize_subscribe_glob(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
    ) -> Optional[str]:
        return None

    async def is_subscribe_glob_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return AuthResult.OK

    async def authorize_notify(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        identifier: bytes,
        message_sha512: bytes,
        now: float,
    ) -> Optional[str]:
        return None

    async def is_notify_allowed(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        identifier: bytes,
        message_sha512: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return AuthResult.OK

    async def authorize_stateful_configure(
        self,
        /,
        *,
        tracing: bytes,
        subscriber_nonce: bytes,
        enable_zstd: bool,
        enable_training: bool,
        initial_dict: int,
    ) -> Optional[str]:
        return None

    async def is_stateful_configure_allowed(
        self, /, *, message: S2B_Configure, now: float
    ) -> AuthResult:
        return AuthResult.OK

    async def authorize_check_subscriptions(
        self, /, *, tracing: bytes, url: str, now: float
    ) -> Optional[str]:
        return None

    async def is_check_subscriptions_allowed(
        self, /, *, tracing: bytes, url: str, now: float, authorization: Optional[str]
    ) -> AuthResult:
        return AuthResult.OK

    async def authorize_set_subscriptions(
        self, /, *, tracing: bytes, url: str, strong_etag: StrongEtag, now: float
    ) -> Optional[str]:
        return None

    async def is_set_subscriptions_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        strong_etag: StrongEtag,
        subscriptions: SetSubscriptionsInfo,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return AuthResult.OK

    async def authorize_stateful_continue_receive(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        part_id: int,
        url: str,
        now: float,
    ) -> Optional[str]:
        return None

    async def is_stateful_continue_receive_allowed(
        self, /, *, url: str, message: S2B_ContinueReceive, now: float
    ) -> AuthResult:
        return AuthResult.OK

    async def authorize_confirm_receive(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        num_subscribers: int,
        url: str,
        now: float,
    ) -> Optional[str]:
        return None

    async def is_confirm_receive_allowed(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        num_subscribers: int,
        url: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return AuthResult.OK

    async def authorize_confirm_missed(
        self, /, *, tracing: bytes, topic: bytes, url: str, now: float
    ) -> Optional[str]:
        return None

    async def is_confirm_missed_allowed(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        url: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return AuthResult.OK


class ToSubscriberNoneAuth:
    """Sets up and allows a subscriber that does not block any incoming requests.

    In order for this to be secure, the subscribers must only be able to receive
    messages from trusted clients.
    """

    async def setup_to_subscriber_auth(self) -> None: ...
    async def teardown_to_subscriber_auth(self) -> None: ...

    async def authorize_receive(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        identifier: bytes,
        now: float,
    ) -> Optional[str]:
        return None

    async def is_receive_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        identifier: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return AuthResult.OK

    async def authorize_missed(
        self, /, *, tracing: bytes, recovery: str, topic: bytes, now: float
    ) -> Optional[str]:
        return None

    async def is_missed_allowed(
        self,
        /,
        *,
        tracing: bytes,
        recovery: str,
        topic: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return AuthResult.OK

    async def authorize_confirm_subscribe_exact(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
    ) -> Optional[str]:
        return None

    async def is_confirm_subscribe_exact_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return AuthResult.OK

    async def authorize_confirm_subscribe_glob(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
    ) -> Optional[str]:
        return None

    async def is_confirm_subscribe_glob_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return AuthResult.OK

    async def authorize_confirm_notify(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        subscribers: int,
        topic: bytes,
        message_sha512: bytes,
        now: float,
    ) -> Optional[str]:
        return None

    async def is_confirm_notify_allowed(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        subscribers: int,
        topic: bytes,
        message_sha512: bytes,
        authorization: Optional[str],
        now: float,
    ) -> AuthResult:
        return AuthResult.OK

    async def authorize_stateful_confirm_configure(
        self, /, *, broadcaster_nonce: bytes, tracing: bytes, now: float
    ) -> Optional[str]:
        return None

    async def is_stateful_confirm_configure_allowed(
        self, /, *, message: B2S_ConfirmConfigure, now: float
    ) -> AuthResult:
        return AuthResult.OK

    async def authorize_stateful_enable_zstd_preset(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        compressor_identifier: int,
        compression_level: int,
        min_size: int,
        max_size: int,
        authorization: Optional[str],
        now: float,
    ) -> Optional[str]:
        return None

    async def is_stateful_enable_zstd_preset_allowed(
        self, /, *, url: str, message: B2S_EnableZstdPreset, now: float
    ) -> AuthResult:
        return AuthResult.OK

    async def authorize_stateful_enable_zstd_custom(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        compressor_identifier: int,
        compression_level: int,
        min_size: int,
        max_size: int,
        sha512: bytes,
        now: float,
    ) -> Optional[str]:
        return None

    async def is_stateful_enable_zstd_custom_allowed(
        self, /, *, url: str, message: B2S_EnableZstdCustom, now: float
    ) -> AuthResult:
        return AuthResult.OK

    async def authorize_stateful_disable_zstd_custom(
        self, /, *, tracing: bytes, compressor_identifier: int, url: str, now: float
    ) -> Optional[str]:
        return None

    async def is_stateful_disable_zstd_custom_allowed(
        self, /, *, url: str, message: B2S_DisableZstdCustom, now: float
    ) -> AuthResult:
        return AuthResult.OK


if TYPE_CHECKING:
    _: Type[ToBroadcasterAuthConfig] = ToBroadcasterNoneAuth
    __: Type[ToSubscriberAuthConfig] = ToSubscriberNoneAuth



from typing import TYPE_CHECKING, Collection, List, Literal, Optional, Type, Union

from lonelypsp.compat import fast_dataclass
from lonelypsp.stateful.constants import (
    PubSubStatefulMessageFlags,
    SubscriberToBroadcasterStatefulMessageType,
)
from lonelypsp.stateful.generic_parser import S2B_MessageParser
from lonelypsp.stateful.parser_helpers import parse_simple_headers
from lonelypsp.stateful.serializer_helpers import (
    MessageSerializer,
    serialize_simple_message,
)
from lonelypsp.sync_io import SyncReadableBytesIO


@fast_dataclass
class S2B_Configure:
    """
    S2B = Subscriber to Broadcaster

    See the type enum documentation for more information on the fields
    """

    type: Literal[SubscriberToBroadcasterStatefulMessageType.CONFIGURE]
    """discriminator value"""

    subscriber_nonce: bytes
    """32 random bytes representing the subscriber's contribution to the nonce"""

    enable_zstd: bool
    """if the client is willing to receive zstandard compressed messages"""

    enable_training: bool
    """if the client may accept custom compression dictionaries"""

    initial_dict: int
    """Either 0 to not recommend an initial preset dictionary, or a positive integer
    representing one of the preset dictionaries that the subscriber believes is appropriate
    for this connection. A value of 1 is ignored.

    Preset dictionaries are typically used when the subscriber may not be connected long
    enough for the cost of training a connection specific dictionary to be properly amortized.
    """

    authorization: Optional[str]
    """the authorization header provided by the subscriber to show the broadcaster it
    is allowed to connect

    empty strings are converted to None for consistency with http endpoints
    """

    tracing: bytes
    """the tracing data for the message, which may be empty"""


_headers: Collection[str] = (
    "x-subscriber-nonce",
    "x-enable-zstd",
    "x-enable-training",
    "x-initial-dict",
    "authorization",
    "x-tracing",
)


class S2B_ConfigureParser:
    """Satisfies S2B_MessageParser[S2B_Configure]"""

    @classmethod
    def relevant_types(cls) -> List[SubscriberToBroadcasterStatefulMessageType]:
        return [SubscriberToBroadcasterStatefulMessageType.CONFIGURE]

    @classmethod
    def parse(
        cls,
        flags: PubSubStatefulMessageFlags,
        type: SubscriberToBroadcasterStatefulMessageType,
        payload: SyncReadableBytesIO,
    ) -> S2B_Configure:
        assert type == SubscriberToBroadcasterStatefulMessageType.CONFIGURE

        headers = parse_simple_headers(flags, payload, _headers)
        subscriber_nonce = headers["x-subscriber-nonce"]
        if len(subscriber_nonce) != 32:
            raise ValueError("x-subscriber-nonce must be 32 bytes")

        enable_zstd = headers["x-enable-zstd"] == b"\x01"
        enable_training = headers["x-enable-training"] == b"\x01"

        initial_dict_bytes = headers.get("x-initial-dict", b"0")
        if len(initial_dict_bytes) > 2:
            raise ValueError("x-initial-dict max 2 bytes")

        initial_dict = int.from_bytes(initial_dict_bytes, "big")
        if initial_dict < 0:
            raise ValueError("x-initial-dict must be non-negative")

        authorization_bytes = headers["authorization"]
        authorization: Optional[str] = None
        if authorization_bytes != b"":
            try:
                authorization = authorization_bytes.decode("utf-8")
            except UnicodeDecodeError:
                raise ValueError("authorization must be a utf-8 string")

        tracing = headers["x-tracing"]

        return S2B_Configure(
            type=type,
            subscriber_nonce=subscriber_nonce,
            enable_zstd=enable_zstd,
            enable_training=enable_training,
            initial_dict=initial_dict,
            authorization=authorization,
            tracing=tracing,
        )


if TYPE_CHECKING:
    _: Type[S2B_MessageParser[S2B_Configure]] = S2B_ConfigureParser


def serialize_s2b_configure(
    msg: S2B_Configure, /, *, minimal_headers: bool
) -> Union[bytes, bytearray]:
    """Satisfies MessageSerializer[S2B_Configure]"""
    return serialize_simple_message(
        type=msg.type,
        header_names=_headers,
        header_values=(
            msg.subscriber_nonce,
            b"\x01" if msg.enable_zstd else b"\x00",
            b"\x01" if msg.enable_training else b"\x00",
            msg.initial_dict.to_bytes(2, "big"),
            msg.authorization.encode("utf-8") if msg.authorization is not None else b"",
            msg.tracing,
        ),
        payload=b"",
        minimal_headers=minimal_headers,
    )


if TYPE_CHECKING:
    __: MessageSerializer[S2B_Configure] = serialize_s2b_configure



from typing import TYPE_CHECKING, List, Optional, Sequence, Type

from lonelypsp.stateful.constants import (
    BroadcasterToSubscriberStatefulMessageType,
    PubSubStatefulMessageFlags,
    SubscriberToBroadcasterStatefulMessageType,
)
from lonelypsp.stateful.generic_parser import B2S_MessageParser, S2B_MessageParser
from lonelypsp.stateful.message import B2S_Message, S2B_Message
from lonelypsp.stateful.messages.configure import S2B_ConfigureParser
from lonelypsp.stateful.messages.confirm_configure import B2S_ConfirmConfigureParser
from lonelypsp.stateful.messages.confirm_notify import B2S_ConfirmNotifyParser
from lonelypsp.stateful.messages.confirm_receive import S2B_ConfirmRecieveParser
from lonelypsp.stateful.messages.confirm_subscribe import (
    B2S_ConfirmSubscribeExactParser,
    B2S_ConfirmSubscribeGlobParser,
)
from lonelypsp.stateful.messages.confirm_unsubscribe import (
    B2S_ConfirmUnsubscribeExactParser,
    B2S_ConfirmUnsubscribeGlobParser,
)
from lonelypsp.stateful.messages.continue_notify import B2S_ContinueNotifyParser
from lonelypsp.stateful.messages.continue_receive import S2B_ContinueReceiveParser
from lonelypsp.stateful.messages.disable_zstd_custom import B2S_DisableZstdCustomParser
from lonelypsp.stateful.messages.enable_zstd_custom import B2S_EnableZstdCustomParser
from lonelypsp.stateful.messages.enable_zstd_preset import B2S_EnableZstdPresetParser
from lonelypsp.stateful.messages.missed import B2S_MissedParser
from lonelypsp.stateful.messages.notify import S2B_NotifyParser
from lonelypsp.stateful.messages.notify_stream import S2B_NotifyStreamParser
from lonelypsp.stateful.messages.receive_stream import B2S_ReceiveStreamParser
from lonelypsp.stateful.messages.subscribe import (
    S2B_SubscribeExactParser,
    S2B_SubscribeGlobParser,
)
from lonelypsp.stateful.messages.unsubscribe import (
    S2B_UnsubscribeExactParser,
    S2B_UnsubscribeGlobParser,
)
from lonelypsp.sync_io import SyncReadableBytesIO

S2B_MESSAGE_PARSERS: List[Type[S2B_MessageParser[S2B_Message]]] = [
    S2B_ConfigureParser,
    S2B_ConfirmRecieveParser,
    S2B_ContinueReceiveParser,
    S2B_NotifyStreamParser,
    S2B_NotifyParser,
    S2B_SubscribeExactParser,
    S2B_SubscribeGlobParser,
    S2B_UnsubscribeExactParser,
    S2B_UnsubscribeGlobParser,
]
"""All of the message parsers for subscriber -> broadcaster messages. Generally,
easier to use via S2B_AnyMessageParser
"""

B2S_MESSAGE_PARSERS: List[Type[B2S_MessageParser[B2S_Message]]] = [
    B2S_ConfirmConfigureParser,
    B2S_ConfirmNotifyParser,
    B2S_ConfirmSubscribeExactParser,
    B2S_ConfirmSubscribeGlobParser,
    B2S_ConfirmUnsubscribeExactParser,
    B2S_ConfirmUnsubscribeGlobParser,
    B2S_ContinueNotifyParser,
    B2S_DisableZstdCustomParser,
    B2S_EnableZstdCustomParser,
    B2S_EnableZstdPresetParser,
    B2S_MissedParser,
    B2S_ReceiveStreamParser,
]
"""All of the message parsers for broadcaster -> subscriber messages. Generally,
easier to use via B2S_AnyMessageParser
"""


def _make_s2b_parser_lookup() -> (
    Sequence[Optional[Type[S2B_MessageParser[S2B_Message]]]]
):

    largest_type: int = 0
    total_types: int = 0
    for parser in S2B_MESSAGE_PARSERS:
        for typ in parser.relevant_types():
            if typ < 0:
                raise ValueError(f"Message type {typ} is negative")
            if typ > largest_type:
                largest_type = typ
            total_types += 1

    assert largest_type < 2 * total_types, "need to swap this implementation"

    lookup: List[Optional[Type[S2B_MessageParser[S2B_Message]]]] = [
        None for _ in range(largest_type + 1)
    ]
    for parser in S2B_MESSAGE_PARSERS:
        for message_type in parser.relevant_types():
            assert (
                lookup[message_type] is None
            ), f"{message_type} has duplicate parser (first: {lookup[message_type]}, second: {parser})"
            lookup[message_type] = parser
    return tuple(lookup)


_s2b_parser_lookup = _make_s2b_parser_lookup()


class S2B_AnyMessageParser:
    """Satisfies S2B_MessageParser[S2B_Message]"""

    @classmethod
    def relevant_types(cls) -> List[SubscriberToBroadcasterStatefulMessageType]:
        result = []
        for parser in S2B_MESSAGE_PARSERS:
            result.extend(parser.relevant_types())
        return result

    @classmethod
    def parse(
        cls,
        flags: PubSubStatefulMessageFlags,
        type: SubscriberToBroadcasterStatefulMessageType,
        payload: SyncReadableBytesIO,
    ) -> S2B_Message:
        if type < 0:
            raise ValueError(f"Message type {type} is negative")
        if type >= len(_s2b_parser_lookup):
            raise ValueError(f"Message type {type} is too large")
        parser = _s2b_parser_lookup[type]
        if parser is None:
            raise ValueError(f"Message type {type} is not supported")
        return parser.parse(flags, type, payload)


if TYPE_CHECKING:
    _: Type[S2B_MessageParser[S2B_Message]] = S2B_AnyMessageParser


def _make_b2s_parser_lookup() -> (
    Sequence[Optional[Type[B2S_MessageParser[B2S_Message]]]]
):
    largest_type: int = 0
    total_types: int = 0
    for parser in B2S_MESSAGE_PARSERS:
        for typ in parser.relevant_types():
            if typ < 0:
                raise ValueError(f"Message type {typ} is negative")
            if typ > largest_type:
                largest_type = typ
            total_types += 1

    assert largest_type < 2 * total_types, "need to swap this implementation"

    lookup: List[Optional[Type[B2S_MessageParser[B2S_Message]]]] = [
        None for _ in range(largest_type + 1)
    ]
    for parser in B2S_MESSAGE_PARSERS:
        for message_type in parser.relevant_types():
            assert (
                lookup[message_type] is None
            ), f"{message_type} has duplicate parser (first: {lookup[message_type]}, second: {parser})"
            lookup[message_type] = parser
    return tuple(lookup)


_b2s_parser_lookup = _make_b2s_parser_lookup()


class B2S_AnyMessageParser:
    """Satisfies B2S_MessageParser[B2S_Message]"""

    @classmethod
    def relevant_types(cls) -> List[BroadcasterToSubscriberStatefulMessageType]:
        result = []
        for parser in B2S_MESSAGE_PARSERS:
            result.extend(parser.relevant_types())
        return result

    @classmethod
    def parse(
        cls,
        flags: PubSubStatefulMessageFlags,
        type: BroadcasterToSubscriberStatefulMessageType,
        payload: SyncReadableBytesIO,
    ) -> B2S_Message:
        if type < 0:
            raise ValueError(f"Message type {type} is negative")
        if type >= len(_b2s_parser_lookup):
            raise ValueError(f"Message type {type} is too large")
        parser = _b2s_parser_lookup[type]
        if parser is None:
            raise ValueError(f"Message type {type} is not supported")
        return parser.parse(flags, type, payload)


if TYPE_CHECKING:
    __: Type[B2S_MessageParser[B2S_Message]] = B2S_AnyMessageParser



from typing import TYPE_CHECKING, Collection, List, Literal, Optional, Type, Union

from lonelypsp.compat import fast_dataclass
from lonelypsp.stateful.constants import (
    BroadcasterToSubscriberStatefulMessageType,
    PubSubStatefulMessageFlags,
)
from lonelypsp.stateful.generic_parser import B2S_MessageParser
from lonelypsp.stateful.parser_helpers import parse_simple_headers
from lonelypsp.stateful.serializer_helpers import (
    MessageSerializer,
    int_to_minimal_unsigned,
    serialize_simple_message,
)
from lonelypsp.sync_io import SyncReadableBytesIO


@fast_dataclass
class B2S_EnableZstdPreset:
    """
    B2S = Broadcaster to Subscriber
    See the type enum documentation for more information on the fields
    """

    type: Literal[BroadcasterToSubscriberStatefulMessageType.ENABLE_ZSTD_PRESET]
    """discriminator value"""

    identifier: int
    """the agreed upon identifier for this preset; 1 always means compression
    without a custom dictionary (i.e., the dictionary is sent alongside the
    compressed data)
    """

    compression_level: int
    """the compression level (any negative integer up to and including positive 22)
    that the broadcaster recommends for this preset; the subscriber is free to
    ignore this recommendation
    """

    min_size: int
    """the minimum in size in bytes that the broadcaster recommends for using
    this preset; the subscriber is free to ignore this recommendation
    """

    max_size: int
    """the maximum in size in bytes that the broadcaster recommends for using
    this preset; the subscriber is free to ignore this recommendation. 2**64-1
    for no limit
    """

    authorization: Optional[str]
    """the authorization header that shows this was sent by the broadcaster

    empty strings are converted to None for consistency with http endpoints
    """

    tracing: bytes
    """the tracing data, which may be empty"""


_headers: Collection[str] = (
    "x-identifier",
    "x-compression-level",
    "x-min-size",
    "x-max-size",
    "authorization",
    "x-tracing",
)


class B2S_EnableZstdPresetParser:
    """Satisfies B2S_MessageParser[B2S_EnableZstdPreset]"""

    @classmethod
    def relevant_types(cls) -> List[BroadcasterToSubscriberStatefulMessageType]:
        return [BroadcasterToSubscriberStatefulMessageType.ENABLE_ZSTD_PRESET]

    @classmethod
    def parse(
        cls,
        flags: PubSubStatefulMessageFlags,
        type: BroadcasterToSubscriberStatefulMessageType,
        payload: SyncReadableBytesIO,
    ) -> B2S_EnableZstdPreset:
        assert type == BroadcasterToSubscriberStatefulMessageType.ENABLE_ZSTD_PRESET

        headers = parse_simple_headers(flags, payload, _headers)
        identifier_bytes = headers["x-identifier"]
        if len(identifier_bytes) > 8:
            raise ValueError("x-identifier must be at most 8 bytes")

        identifier = int.from_bytes(identifier_bytes, "big")

        compression_level_bytes = headers["x-compression-level"]
        if len(compression_level_bytes) > 2:
            raise ValueError("x-compression-level max 2 bytes")

        compression_level = int.from_bytes(compression_level_bytes, "big", signed=True)

        min_size_bytes = headers["x-min-size"]
        if len(min_size_bytes) > 4:
            raise ValueError("x-min-size max 4 bytes")

        min_size = int.from_bytes(min_size_bytes, "big")

        max_size_bytes = headers["x-max-size"]
        if len(max_size_bytes) > 8:
            raise ValueError("x-max-size max 8 bytes")

        max_size = int.from_bytes(max_size_bytes, "big")

        authorization_bytes = headers["authorization"]
        authorization: Optional[str] = None
        if authorization_bytes != b"":
            try:
                authorization = authorization_bytes.decode("utf-8")
            except UnicodeDecodeError:
                raise ValueError("authorization must be a utf-8 string")

        tracing = headers["x-tracing"]

        return B2S_EnableZstdPreset(
            type=type,
            identifier=identifier,
            compression_level=compression_level,
            min_size=min_size,
            max_size=max_size,
            authorization=authorization,
            tracing=tracing,
        )


if TYPE_CHECKING:
    _: Type[B2S_MessageParser[B2S_EnableZstdPreset]] = B2S_EnableZstdPresetParser


def serialize_b2s_enable_zstd_preset(
    msg: B2S_EnableZstdPreset, /, *, minimal_headers: bool
) -> Union[bytes, bytearray]:
    """Satisfies MessageSerializer[B2S_EnableZstdPreset]"""
    return serialize_simple_message(
        type=msg.type,
        header_names=_headers,
        header_values=(
            int_to_minimal_unsigned(msg.identifier),
            msg.compression_level.to_bytes(2, "big", signed=True),
            int_to_minimal_unsigned(msg.min_size),
            int_to_minimal_unsigned(msg.max_size),
            msg.authorization.encode("utf-8") if msg.authorization is not None else b"",
            msg.tracing,
        ),
        payload=b"",
        minimal_headers=minimal_headers,
    )


if TYPE_CHECKING:
    __: MessageSerializer[B2S_EnableZstdPreset] = serialize_b2s_enable_zstd_preset



# ruff: noqa: T201
"""fuzzes the bounded queue behaving like a list"""

import random
import sys
from typing import List

from lonelypsp.util.bounded_deque import BoundedDeque


def _fuzz() -> None:
    try:
        for i in range(10000):
            if i % 100 == 0:
                print("." if i % 1000 != 0 else "o", end="")
                sys.stdout.flush()
            seed = random.randint(0, 2**64 - 1)
            _test_seed(seed)
    except BaseException:
        print()
        raise

    print("done")


def _test_seed(seed: int) -> None:
    random.seed(seed)

    testing: BoundedDeque[int] = BoundedDeque()
    good: List[int] = []

    ctr = 0
    for i in range(1000):
        choice = random.choice(
            ["remove", "pop", "popleft"]
            if len(good) > 1000
            else (
                ["append", "appendleft", "remove", "pop", "popleft"]
                if good
                else ["append", "appendleft"]
            )
        )

        if choice == "append":
            testing.append(ctr)
            good.append(ctr)
            ctr += 1
        elif choice == "appendleft":
            testing.appendleft(ctr)
            good.insert(0, ctr)
            ctr += 1
        elif choice == "pop":
            testing.pop()
            good.pop()
        elif choice == "popleft":
            testing.popleft()
            good.pop(0)
        elif choice == "remove":
            to_remove = random.choice(good)
            testing.remove(to_remove)
            good.remove(to_remove)

        assert (
            list(testing) == good
        ), f"seed={seed} on iteration={i} has {testing} vs {good=}"
        # verify all unused parts are None

        real_idx = testing.length
        array_idx = (testing.start + testing.length) % len(testing.data)
        capacity = len(testing.data)
        while real_idx < capacity:
            assert (
                testing.data[array_idx] is None
            ), f"seed={seed} on iteration={i} has corrupt {testing}, {testing.data=}"
            real_idx += 1
            array_idx += 1
            if array_idx == capacity:
                array_idx = 0


if __name__ == "__main__":
    _fuzz()



from typing import Generic, TypeVar

from lonelypsp.compat import fast_dataclass

T = TypeVar("T")


@fast_dataclass
class TracingAndFollowup(Generic[T]):
    tracing: bytes
    """the tracing data to transfer"""

    followup: T
    """the object that will pick up tracing (on this side) after the response"""



from typing import Optional

from lonelypsp.compat import fast_dataclass


@fast_dataclass
class SimpleTracingConfig:
    """Configuration for the simple tracing implementation"""

    trace_max_age_seconds: Optional[int] = 60 * 60 * 24 * 90
    """The maximum age of traces before they are cleaned up by the next run; None for
    no maximum age
    """



import hashlib
import io
from typing import List, Literal, Optional

from lonelypsp.compat import fast_dataclass


@fast_dataclass
class StrongEtag:
    format: Literal[0]
    """reserved discriminator value"""

    etag: bytes
    """the SHA512 hash of the document"""


@fast_dataclass
class GlobAndRecovery:
    glob: str
    recovery: Optional[str]


@fast_dataclass
class TopicAndRecovery:
    topic: bytes
    recovery: Optional[str]


def _encode_recovery(recovery: Optional[str]) -> bytes:
    if recovery is None:
        return b""
    return recovery.encode("utf-8")


class _PreallocatedBytesIO:
    def __init__(self, size: int) -> None:
        self.buf = bytearray(size)
        self.pos = 0

    def write(self, data: bytes) -> None:
        assert self.pos + len(data) <= len(self.buf)
        self.buf[self.pos : self.pos + len(data)] = data
        self.pos += len(data)

    def getbuf(self) -> bytearray:
        assert self.pos == len(self.buf)
        return self.buf


def make_strong_etag(
    url: str,
    topics: List[TopicAndRecovery],
    globs: List[GlobAndRecovery],
    *,
    recheck_sort: bool = True
) -> StrongEtag:
    """Generates the strong etag for `CHECK_SUBSCRIPTIONS` and
    `SET_SUBSCRIPTIONS` in a single pass; this is useful for reference
    or when there are a small number of topics/globs, but the etag
    can be generated from streaming data using `create_strong_etag_generator`

    NOTE: the topics and globs MUST be in (bytewise) lexicographic order
    for this to be stable. If `recheck_sort` is `True`, this will raise
    `ValueError` if the topics or globs are not sorted properly. If
    explicitly set to `False`, then the caller must have already ensured
    the topics and globs are sorted properly
    """

    if recheck_sort:
        for idx, topic in enumerate(topics):
            if idx > 0 and topic.topic <= topics[idx - 1].topic:
                raise ValueError(
                    "topics must be unique and in ascending lexicographic order"
                )

        for idx, glob in enumerate(globs):
            if idx > 0 and glob.glob <= globs[idx - 1].glob:
                raise ValueError(
                    "globs must be unique and in ascending lexicographic order"
                )

    doc = io.BytesIO()
    doc.write(b"URL")

    encoded_url = url.encode("utf-8")
    doc.write(len(encoded_url).to_bytes(2, "big"))
    doc.write(encoded_url)

    doc.write(b"\nEXACT")
    for topic in topics:
        doc.write(len(topic.topic).to_bytes(2, "big"))
        doc.write(topic.topic)
        encoded_recovery = _encode_recovery(topic.recovery)
        doc.write(len(encoded_recovery).to_bytes(2, "big"))
        doc.write(encoded_recovery)

    doc.write(b"\nGLOB")
    for glob in globs:
        encoded_glob = glob.glob.encode("utf-8")
        doc.write(len(encoded_glob).to_bytes(2, "big"))
        doc.write(encoded_glob)
        encoded_recovery = _encode_recovery(glob.recovery)
        doc.write(len(encoded_recovery).to_bytes(2, "big"))
        doc.write(encoded_recovery)

    doc.write(b"\n")
    etag = hashlib.sha512(doc.getvalue()).digest()
    return StrongEtag(format=0, etag=etag)


class StrongEtagGeneratorAtGlobs:
    """Adds glob patterns to the strong etag, then call finish() to get the strong etag"""

    def __init__(self, hasher: "hashlib._Hash", *, recheck_sort: bool = True) -> None:
        self.hasher = hasher
        self._recheck_sort = recheck_sort
        self._last_glob: Optional[bytes] = None

    def add_glob(self, *globs: GlobAndRecovery) -> "StrongEtagGeneratorAtGlobs":
        """Add the given glob or globs to the strong etag; multiple globs can be
        faster than calling add_glob multiple times as it reduces calls to the
        underlying hasher's update method, but requires more memory
        """
        if len(globs) == 0:
            return self

        encoded_globs = [g.glob.encode("utf-8") for g in globs]

        if self._recheck_sort:
            for g in encoded_globs:
                if self._last_glob is not None and g <= self._last_glob:
                    raise ValueError(
                        "globs must be unique and in ascending lexicographic order"
                    )
                self._last_glob = g

        encoded_recoveries = [_encode_recovery(g.recovery) for g in globs]

        buf = _PreallocatedBytesIO(
            4 * len(globs)
            + sum(len(g) for g in encoded_globs)
            + sum(len(r) for r in encoded_recoveries)
        )
        for encoded_glob, encoded_recovery in zip(encoded_globs, encoded_recoveries):
            buf.write(len(encoded_glob).to_bytes(2, "big"))
            buf.write(encoded_glob)
            buf.write(len(encoded_recovery).to_bytes(2, "big"))
            buf.write(encoded_recovery)

        self.hasher.update(buf.getbuf())
        return self

    def finish(self) -> StrongEtag:
        self.hasher.update(b"\n")
        return StrongEtag(format=0, etag=self.hasher.digest())


class StrongEtagGeneratorAtTopics:
    """Adds topics to the strong etag, then call finish_topics() to move onto globs"""

    def __init__(self, hasher: "hashlib._Hash", *, recheck_sort: bool = True) -> None:
        self.hasher = hasher
        self._recheck_sort = recheck_sort
        self._last_topic: Optional[bytes] = None

    def add_topic(self, *topic: TopicAndRecovery) -> "StrongEtagGeneratorAtTopics":
        """Add the given topic or topics to the strong etag; multiple topics can be
        faster than calling add_topic multiple times as it reduces calls to the
        underlying hasher's update method, but requires more memory
        """
        if len(topic) == 0:
            return self

        if self._recheck_sort:
            for t in topic:
                if self._last_topic is not None and t.topic <= self._last_topic:
                    raise ValueError(
                        "topics must be unique and in ascending lexicographic order"
                    )
                self._last_topic = t.topic

        encoded_recoveries = [_encode_recovery(t.recovery) for t in topic]

        buf = _PreallocatedBytesIO(
            4 * len(topic)
            + sum(len(t.topic) for t in topic)
            + sum(len(r) for r in encoded_recoveries)
        )
        for t, r in zip(topic, encoded_recoveries):
            buf.write(len(t.topic).to_bytes(2, "big"))
            buf.write(t.topic)
            buf.write(len(r).to_bytes(2, "big"))
            buf.write(r)

        self.hasher.update(buf.getbuf())
        return self

    def finish_topics(self) -> StrongEtagGeneratorAtGlobs:
        self.hasher.update(b"\nGLOB")
        return StrongEtagGeneratorAtGlobs(self.hasher, recheck_sort=self._recheck_sort)


def create_strong_etag_generator(
    url: str, *, recheck_sort: bool = True
) -> StrongEtagGeneratorAtTopics:
    """Returns a StrongEtagGeneratorAtTopics that can be used to add topics and
    globs to the strong etag, then call finish_topics() to get the generator for
    adding globs, then call finish() to get the strong etag. This avoids having
    to ever store the actual document being hashed but requires more calls to
    the underlying hasher's update method

    Example usage:

    ```python
    etag = (
        create_strong_etag_generator("https://example.com", recheck_sort=False)
        .add_topic(
            TopicAndRecovery(b"topic1", "recovery1"),
            TopicAndRecovery(b"topic2", "recovery2")
        )
        .finish_topics()
        .add_glob(
            GlobAndRecovery("glob1", "recovery3"),
            GlobAndRecovery("glob2", "recovery4")
        )
        .finish()
    )
    ```

    NOTE: the topics and globs MUST be in (bytewise) lexicographic order
    for this to be stable. If `recheck_sort` is `True`, this will raise
    `ValueError` if the topics or globs are not sorted properly. If
    explicitly set to `False`, then the caller must have already ensured
    the topics and globs are sorted properly
    """
    encoded_url = url.encode("utf-8")
    buf = _PreallocatedBytesIO(3 + 2 + len(encoded_url) + 6)

    buf.write(b"URL")
    buf.write(len(encoded_url).to_bytes(2, "big"))
    buf.write(encoded_url)
    buf.write(b"\nEXACT")

    hasher = hashlib.sha512(buf.getbuf())
    return StrongEtagGeneratorAtTopics(hasher, recheck_sort=recheck_sort)



from typing import Generic, Protocol, TypeVar

from lonelypsp.tracing.stateless.root import (
    StatelessTracingBroadcasterRoot,
    StatelessTracingSubscriberRoot,
)

InitializerTcontra = TypeVar("InitializerTcontra", contravariant=True)
"""When using OpenTelemetry style spans, the type that allows the caller to
embed the span within some other operation. We never pass this in a keyword
argument (and enforce this with the protocol) so you can use a better name
for it
"""


class TracingBroadcasterRoot(Generic[InitializerTcontra], Protocol):
    """
    Can produce stateless or stateful roots from the broadcaster
    perspective for tracing. To avoid excessive garbage collection
    it is not intended that this produces new objects on each
    access.

    Doesn't absorb the two protocols as it is convenient for them to
    have overlapping names
    """

    @property
    def stateless(self) -> "StatelessTracingBroadcasterRoot[InitializerTcontra]":
        """Returns the stateless root"""

    # @property
    # def stateful(self) -> "StatefulTracingBroadcasterRoot[InitializerTcontra]":
    #     """Produces the stateful root"""


class TracingSubscriberRoot(Generic[InitializerTcontra], Protocol):
    """
    Can produce stateless or stateful roots from the subscriber perspective for
    tracing. To avoid excessive garbage collection it is not intended that this
    produces new objects on each access.

    Doesn't absorb the two protocols as it is convenient for them to have
    overlapping names
    """

    @property
    def stateless(self) -> "StatelessTracingSubscriberRoot[InitializerTcontra]":
        """Returns the stateless root"""

    # @property
    # def stateful(self) -> "StatefulTracingSubscriberRoot[InitializerTcontra]":
    #     """Produces the stateful root"""



import sys
from dataclasses import dataclass as fast_dataclass
from functools import partial
from typing import TypeVar

__all__ = ["fast_dataclass", "assert_never"]

T = TypeVar("T")
if sys.version_info >= (3, 10):

    def make_dataclass_fast(x: T) -> T:
        return partial(x, frozen=True, slots=True)  # type: ignore

    fast_dataclass = make_dataclass_fast(fast_dataclass)
else:

    def make_dataclass_fast(x: T) -> T:
        return partial(x, frozen=True)  # type: ignore

    fast_dataclass = make_dataclass_fast(fast_dataclass)


if sys.version_info >= (3, 11):
    from typing import assert_never
else:
    from typing import NoReturn
    from typing import NoReturn as Never

    def assert_never(value: Never) -> NoReturn:
        raise AssertionError(f"Unhandled type: {value!r}")



import io

from fastapi import APIRouter

from lonelypsp.tracing.impl.simple.db import SimpleTracingDB

router = APIRouter()


DB_PATH: str = "tracing.db"


@router.get("/index.html")
def index() -> str:
    result = io.StringIO()
    result.write(
        """
<html>
<head>
    <title>LonelyPS - stats</title>
</head>
<body>
"""
    )

    with SimpleTracingDB(DB_PATH) as db:
        assert db.cursor is not None
        result.write("<h1>Notify</h1>")

        db.cursor.execute("BEGIN DEFERRED TRANSACTION")
        try:
            db.cursor.execute("SELECT COUNT(*) FROM stateless_notifies")
            num_notifies = db.cursor.fetchone()[0]

            result.write(f"<p>There were {num_notifies} notifications to ")

            db.cursor.execute(
                "SELECT COUNT(DISTINCT json_extract(extra, '$.broadcaster')) "
                "FROM stateless_notify_timings "
                "WHERE name = 'sending_request'"
            )
            num_broadcaster = db.cursor.fetchone()[0]
            result.write(f"{num_broadcaster} broadcasters</p>")

            result.write("</p>")
        finally:
            db.cursor.execute("COMMIT TRANSACTION")

    result.write("</body></html>")
    return result.getvalue()



from typing import TYPE_CHECKING, Type

from lonelypsp.tracing.impl.simple.config import SimpleTracingConfig
from lonelypsp.tracing.impl.simple.db import SimpleTracingDBSidecar
from lonelypsp.tracing.impl.simple.stateless.root import (
    SimpleStatelessTracingBroadcasterRoot,
    SimpleStatelessTracingSubscriberRoot,
)
from lonelypsp.tracing.root import TracingBroadcasterRoot, TracingSubscriberRoot


class SimpleTracingBroadcasterRoot:
    def __init__(self) -> None:
        self.stateless = SimpleStatelessTracingBroadcasterRoot()


class SimpleTracingSubscriberRoot:
    def __init__(self, db: SimpleTracingDBSidecar, config: SimpleTracingConfig) -> None:
        self.stateless = SimpleStatelessTracingSubscriberRoot(db, config)


if TYPE_CHECKING:
    _: Type[TracingBroadcasterRoot] = SimpleTracingBroadcasterRoot
    __: Type[TracingSubscriberRoot] = SimpleTracingSubscriberRoot



from enum import IntEnum, IntFlag, auto


class PubSubStatefulMessageFlags(IntFlag):
    # Examples of flags we might want to add in the future, should the need arise:
    # - huge headers (2 -> 4 bytes for length, allows for huge topic names)
    # - extensions (e.g., alternative compression algorithms)
    # - forward support flags / backward support flags (as alt. to expanded headers)
    #   forward support meaning enable an experimental feature
    #   backward support meaning enable a removed feature
    MINIMAL_HEADERS = 1 << 0
    """Our websocket message bodies include a headers section. This is not
    related to the headers within the HTTP / WS protocol. Our headers take one
    of two formats, depending on if this flag is set or not.

    If this flag is set, the broadcaster and subscriber agree on the headers
    that should be sent in each packet, and only send the values.

    If this flag is not set, the sender expresses the headers with (key, value)
    pairs, allowing more flexibility for a version mismatch but also more
    overhead.

    Typically it's helpful to turn this off when in the process of updating the
    server and clients, then turn it on when all broadcasters and subscribers
    are on the same version.

    Individual header names and values are always restricted to no more than
    65535 bytes.
    """


class SubscriberToBroadcasterStatefulMessageType(IntEnum):
    """Assigns a unique integer to each type of message that a subscriber can
    send to a broadcaster over the stateful protocol
    """

    CONFIGURE = auto()
    """The first packet sent over the wire. Provides additional information to
    the broadcaster for how to handle the connection, as well as facilitating
    the generation of a shared public key, required for some authorization
    protocols. This shared public key is referred to as a `connection nonce`
    or just `nonce` in the future and is generated by 
    `SHA256(subscriber_nonce CONCAT broadcaster_nonce)`

    
    ### headers 
    - x-subscriber-nonce: 32 random bytes representing the subscriber's contribution
        to the nonce. The broadcaster will provide its contribution in the response.
    - x-enable-zstd: 1 byte, big-endian, unsigned. 0 to disable zstandard compression,
        1 to indicate the client is willing to receive zstandard compressed messages.
    - x-enable-training: 1 byte, big-endian, unsigned. 0 to indicate the client will not
      accept custom compression dictionaries, 1 to indicate the client may accept them.
    - x-initial-dict: 2 bytes, big-endian, unsigned. 0 to indicate the client does not
      have a specific preset dictionary in mind to use, otherwise, the id of the preset
      dictionary the client thinks is a good fit for this connection
    - authorization: variable length, utf-8 encoded, verifies this subscriber is allowed
      to subscribe via websocket in this manner
    - x-tracing: variable length tracing data, see `Tracing` in readme

    ### body
    none
    """

    SUBSCRIBE_EXACT = auto()
    """Indicates the subscriber wants to receive messages on a specific topic. If
    already subscribed, the broadcaster will disconnect the subscriber.
    
    ### headers 
    - authorization (url: websocket:<nonce>:<ctr>) variable length, utf-8 encoded
    - x-tracing: variable length tracing data, see `Tracing` in readme
    - x-topic: the topic to subscribe to

    ### body
    none
    """

    SUBSCRIBE_GLOB = auto()
    """Indicates the subscriber wants to receive messages to any topic matching the pattern. If
    already subscribed to this exact glob, the broadcaster will disconnect the subscriber.
    
    ### headers
    - authorization (url: websocket:<nonce>:<ctr>) variable length, utf-8 encoded
    - x-tracing: variable length tracing data, see `Tracing` in readme
    - x-glob: the glob pattern to subscribe to (must be valid utf-8)

    ### body
    none
    """

    UNSUBSCRIBE_EXACT = auto()
    """Undoes the effect of a previous SUBSCRIBE_EXACT message. If not subscribed,
    the broadcaster will disconnect the subscriber.
    
    ### headers 
    - authorization (url: websocket:<nonce>:<ctr>) variable length, utf-8 encoded
    - x-tracing: variable length tracing data, see `Tracing` in readme
    - x-topic: the topic to unsubscribe from

    ### body
    none
    """

    UNSUBSCRIBE_GLOB = auto()
    """Undoes the effect of a previous SUBSCRIBE_GLOB message. If not subscribed,
    the broadcaster will disconnect the subscriber.

    ### headers
    - authorization (url: websocket:<nonce>:<ctr>) variable length, utf-8 encoded
    - x-tracing: variable length tracing data, see `Tracing` in readme
    - x-glob: the glob pattern to unsubscribe from

    ### body
    none
    """

    NOTIFY = auto()
    """Informs the broadcaster of a new message to be sent to subscribers of a 
    particular topic. The entire message must go within a single websocket frame
    for this message type; use NOTIFY_STREAM for larger messages

    
    ### headers
    - authorization (url: websocket:<nonce>:<ctr>, see below) variable length, utf-8 encoded
    - x-tracing: variable length tracing data, see `Tracing` in readme
    - x-identifier identifies the notification so the broadcaster can confirm it
    - x-topic is the topic of the notification
    - x-compressor is a big-endian unsigned integer representing one of the previous
      compression methods enabled by the broadcaster
    - x-compressed-length the total length of the compressed body, big-endian, unsigned, max 8 bytes
    - x-decompressed-length the total length of the decompressed body, big-endian, unsigned, max 8 bytes
    - x-compressed-sha512 the sha-512 hash of the compressed content once all parts are concatenated, 64 bytes

    ### body

    the message to send. must have the same hash as the provided hash
    """

    NOTIFY_STREAM = auto()
    """Like notify, asks the broadcaster to send a message to subscribers of a
    particular topic, but allows the message to be split into multiple websocket
    messages (which are themselves potentially split into multiple frames).

    Splitting websocket messages is theoretically redundant as it's already
    handled by the protocol, but in practice websocket message sizes are limited
    to relatively small values (e.g., 16mb) as each step in the process is expecting
    to hold them in memory (often resulting in multiple copies at once)

    NOTE: Notifications may never be weaved. Each notify must complete normally before
    the next can be started or the broadcaster will disconnect the subscriber.

    ### headers

    The following 3 headers are always required, and when using minimal headers, they
    always appear first and in this order:

    - authorization (url: websocket:<nonce>:<ctr>) variable length, utf-8 encoded
    - x-tracing: variable length tracing data, see `Tracing` in readme
    - x-identifier relates the different websocket messages. arbitrary blob, max 64 bytes
    - x-part-id starts at 0 and increments by 1 for each part. unsigned, big-endian, max 8 bytes
    
    If `x-part-id` is 0, the following headers are also required, and when using minimal headers,
    they always appear after the first 3 headers and in this order:

    - x-topic the topic of the notification
    - x-compressor identifiers which compressor (if any) was used for the body. 0 for no compression.
      big-endian, unsigned, max 8 bytes
    - x-compressed-length the total length of the compressed body, big-endian, unsigned, max 8 bytes
    - x-decompressed-length the total length of the decompressed body, big-endian, unsigned, max 8 bytes
    - x-compressed-sha512 the sha-512 hash of the compressed content once all parts are concatenated, 64 bytes

    ### body

    blob of data to append to the compressed notification body
    """

    CONTINUE_RECEIVE = auto()
    """Indicates that the subscriber is ready to receive more data for a notification

    ### headers
    - x-identifier: the identifier of the notification the subscriber needs more parts for
    - x-part-id: the part id that the subscriber received up to, big-endian, unsigned, max 8 bytes
    - authorization: (url: websocket:<nonce>:<ctr>) variable length, utf-8 encoded
    - x-tracing: variable length tracing data, see `Tracing` in readme

    ### body
    none
    """

    CONFIRM_RECEIVE = auto()
    """Indicates that the subscriber received the notification and finished processing it.

    ### headers
    - x-identifier: the identifier of the notification that was received
    - authorization: (url: websocket:<nonce>:<ctr>) variable length, utf-8 encoded
    - x-tracing: variable length tracing data, see `Tracing` in readme
    - x-num-subscribers: the number of subscribers that received the notification, 
        big-endian, unsigned, max 4 bytes. typically 1, but can be more if the subscriber
        forwarded the message to other subscribers

    ### body
    none
    """


class BroadcasterToSubscriberStatefulMessageType(IntEnum):
    """Assigns a unique integer to each type of message that a broadcaster can
    send to a subscriber.
    """

    CONFIRM_CONFIGURE = auto()
    """Indicates that the broadcaster has finished processing a configure message
    sent by the subscriber. Also includes the broadcasters contribution to the
    connections shared public key, required for some authorization protocols.

    ### headers
    - `x-broadcaster-nonce`: (32 bytes)
        the broadcasters contribution for random bytes to the nonce.
        the connection nonce is SHA256(subscriber_nonce CONCAT broadcaster_nonce),
        which is used in the url for generating the authorization header
        when the broadcaster sends a notification to the receiver over
        this websocket and when the subscriber subscribers to a topic over
        this websocket.

        the url is of the form `websocket:<nonce>:<ctr>`, where the ctr is
        a signed 8-byte integer that starts at 1 (or -1) and that depends on if it
        was sent by the broadcaster or subscriber. Both the subscriber and
        broadcaster keep track of both counters; the subscribers counter
        is always negative and decremented by 1 after each subscribe or unsubscribe
        request, the broadcasters counter is always positive and incremented by 1 after
        each notification sent. The nonce is base64url encoded, the ctr is
        hex encoded without a leading 0x and unpadded, e.g.,
        `websocket:abc123:10ffffffffffffff` or `websocket:abc123:-1a`. note that
        the counter changes every time an authorization header is provided,
        even within a single "operation", so e.g. a Notify Stream message broken
        into 6 parts will change the counter 6 times.
    - authorization: variable length, utf-8 encoded, verifies this broadcaster
      is who the subscriber expects
    - x-tracing: variable length tracing data, see `Tracing` in readme
    
    ### body

    none
    """

    CONFIRM_SUBSCRIBE_EXACT = auto()
    """
    Indicates that the subscriber will receive notifications for the given topic

    ### headers
    - x-topic: the topic that the subscriber is now subscribed to
    - authorization: (url: websocket:<nonce>:<ctr>) variable length, utf-8 encoded
    - x-tracing: variable length tracing data, see `Tracing` in readme

    ### body
    none
    """

    CONFIRM_SUBSCRIBE_GLOB = auto()
    """
    Indicates that the subscriber will receive notifications for topics that match
    the given glob pattern

    NOTE: 
        Glob patterns are considered independently. If you subscribe to foo/* and foo/bar/*,
        you will receive a confirmation of foo/* followed by foo/bar/*. On websockets,
        the notifications will be deduplicated so you will only receive one RECEIVE for
        `foo/bar/123` in this case.

    ### headers
    - x-glob: the glob pattern that the subscriber is now subscribed to
    - authorization: (url: websocket:<nonce>:<ctr>) variable length, utf-8 encoded
    - x-tracing: variable length tracing data, see `Tracing` in readme

    ### body
    none
    """

    CONFIRM_UNSUBSCRIBE_EXACT = auto()
    """
    Indicates that the subscriber will no longer receive notifications for the given topic

    ### headers
    - x-topic: the topic that the subscriber is now unsubscribed from
    - authorization: (url: websocket:<nonce>:<ctr>) variable length, utf-8 encoded
    - x-tracing: variable length tracing data, see `Tracing` in readme

    ### body
    none
    """

    CONFIRM_UNSUBSCRIBE_GLOB = auto()
    """
    Indicates that the subscriber will no longer receive notifications for topics that match
    the given glob pattern.

    NOTE: Glob patterns are considered independently

    ### headers
    - x-glob: the glob pattern that the subscriber is now unsubscribed from
    - authorization: (url: websocket:<nonce>:<ctr>) variable length, utf-8 encoded
    - x-tracing: variable length tracing data, see `Tracing` in readme

    ### body
    none
    """

    CONFIRM_NOTIFY = auto()
    """
    Indicates that the broadcaster received the notification and finished processing it.
    This is sent in response to NOTIFY messages and the last NOTIFY STREAM message of
    a notification.

    ### headers
    - x-identifier: the identifier of the notification that was sent
    - x-subscribers: the number of subscribers that received the notification, big-endian,
        unsigned, max 8 bytes
    - authorization: (url: websocket:<nonce>:<ctr>) variable length, utf-8 encoded
    - x-tracing: variable length tracing data, see `Tracing` in readme

    ### body
    none
    """

    CONTINUE_NOTIFY = auto()
    """
    Sent in response to a NOTIFY_STREAM message to indicate that the broadcaster
    is expecting more data for the notification.

    ### headers
    - x-identifier: the identifier of the notification the broadcaster needs more parts for
    - x-part-id: the part id that broadcaster received up to, big-endian, unsigned, max 8 bytes
    - authorization: (url: websocket:<nonce>:<ctr>) variable length, utf-8 encoded
    - x-tracing: variable length tracing data, see `Tracing` in readme

    ### body
    none
    """

    RECEIVE_STREAM = auto()
    """Used to tell the subscriber about a message on a topic it is subscribed to,
    possibly over multiple websocket messages.

    NOTE: Notifications may never be weaved. Each notify must complete normally before
    the next can be started or the subscriber will disconnect. 

    ### headers

    The following 3 headers are always required, and when using minimal headers, they
    always appear first and in this order:

    - authorization (url: websocket:<nonce>:<ctr>)
    - x-tracing: variable length tracing data, see `Tracing` in readme
    - x-identifier relates the different websocket messages. arbitrary blob, max 64 bytes
    - x-part-id starts at 0 and increments by 1 for each part. unsigned, big-endian, max 8 bytes
    
    If `x-part-id` is 0, the following headers are also required, and when using minimal headers,
    they always appear after the first 3 headers and in this order:

    - x-topic the topic of the notification
    - x-compressor identifiers which compressor (if any) was used for the body. 0 for no compression.
      big-endian, unsigned, max 8 bytes
    - x-compressed-length the total length of the compressed body, big-endian, unsigned, max 8 bytes
    - x-decompressed-length the total length of the decompressed body, big-endian, unsigned, max 8 bytes
    - x-compressed-sha512 the sha-512 hash of the compressed content once all parts are concatenated, 64 bytes

    ### body
    blob of data to append to the compressed notification body
    """

    ENABLE_ZSTD_PRESET = auto()
    """Indicates that the broadcaster may start sending, and will accept, data compressed
    using the given preset dictionary. The special dictionary id `1` is reserved to mean
    compression without a custom dictionary (i.e., the dictionary is sent alongside the
    data), and `0` is reserved to mean no compression and is never used in this message.

    A preset dictionary is one with, through some other means than this protocol, the
    broadcasters and subscribers can both agree on the contents of the dictionary without
    actually sending any data. This typically means they just have them on their local disk

    ### headers
    - x-identifier: which compressor is enabled, unsigned, big-endian, max 2 bytes, min 1.
        A value of 1 means compression without a custom dictionary.
    - x-compression-level: what compression level the broadcaster will use with
        this dictionary. signed, big-endian, max 2 bytes, max 22. the subscriber
        is free to choose a different compression level
    - x-min-size: 4 bytes, big-endian, unsigned. a hint to the subscriber for the smallest
        payload the broadcaster will apply this compressor to. the subscriber can use this
        compressor on smaller messages if it wants
    - x-max-size: 8 bytes, big-endian, unsigned. a hint to the susbcriber for the largest
        payload for which the broadcaster will use this compressor. uses 2**64-1 to indicate
        no upper bound. the client can use this compressor on larger messages if it wants
    - authorization (url: websocket:<nonce>:<ctr>)
    - x-tracing: variable length tracing data, see `Tracing` in readme

    ### body
    none
    """

    ENABLE_ZSTD_CUSTOM = auto()
    """Indicates that the broadcaster may start sending, and will accept, data compressed
    using a dictionary which it will send along with this message. The broadcaster generated
    this dictionary by training on the data that has been passed along the websocket so far.
    This dictionary is intended to utilize very specific properties of the messages as they
    are right now; for example, if timestamps are often included in the message in JSON, they 
    will probably all have the same starting prefix which will likely be in the dictionary.

    The broadcaster will retrain the dictionary with some regularity so it can continue to
    take advantage of the most recent data.

    The broadcaster will only accept at most the last 2 custom dictionaries and will only use
    the most recent one.

    ### headers
    - x-identifier: the identifier assigned to the compressor formed with this dictionary.
        unsigned, big-endian, max 8 bytes, min 65536
    - x-compression-level: what compression level the broadcaster will use with
        this dictionary. signed, big-endian, max 2 bytes, max 22. the subscriber
        is free to choose a different compression level
    - x-min-size: max 4 bytes, big-endian, unsigned. a hint to the subscriber for the smallest
        payload the broadcaster will apply this compressor to. the subscriber can use this
        compressor on smaller messages if it wants
    - x-max-size: max 8 bytes, big-endian, unsigned. a hint to the subscriber for the largest
        payload for which the broadcaster will use this compressor. uses 2**64-1 to indicate
        no upper bound. the client can use this compressor on larger messages if it wants.
    - authorization (url: websocket:<nonce>:<ctr>)
    - x-tracing: variable length tracing data, see `Tracing` in readme
    - x-sha512: the sha-512 hash of the dictionary

    ### body
    the dictionary to use for compression.
    """

    DISABLE_ZSTD_CUSTOM = auto()
    """Indicates that the broadcaster will no longer decompress messages using a specific 
    dictionary and that it will no longer send messages compressed with a given dictionary.
    
    ### headers
    - x-identifier: the identifier assigned to the compressor formed with this dictionary.
        unsigned, big-endian, max 8 bytes, min 65536
    - authorization (url: websocket:<nonce>:<ctr>)
    - x-tracing: variable length tracing data, see `Tracing` in readme
    
    ### body
    none
    """

    MISSED = auto()
    """Indicates that the broadcaster may have failed to send some messages on relevant
    topics to this subscriber. This happens when there are multiple broadcasters and a
    relevant message was posted to a different broadcaster than the one this websocket is
    connected with, and that other broadcaster failed to reach this broadcaster.

    ### headers
    - authorization (recovery: websocket:<nonce>:<ctr>)
    - x-tracing: variable length tracing data, see `Tracing` in readme
    - x-topic: the topic that the subscriber may have missed messages on

    ### body
    none
    """



from typing import TYPE_CHECKING, Literal, Type

from lonelypsp.tracing.impl.simple.config import SimpleTracingConfig
from lonelypsp.tracing.impl.simple.db import SimpleTracingDBSidecar
from lonelypsp.tracing.impl.simple.stateless.notify import (
    SimpleStatelessTracingNotifyOnReceived,
    SimpleStatelessTracingNotifyStart,
)
from lonelypsp.tracing.stateless.notify import (
    StatelessTracingNotifyOnReceived,
    StatelessTracingNotifyStart,
)
from lonelypsp.tracing.stateless.root import (
    StatelessTracingBroadcasterRoot,
    StatelessTracingSubscriberRoot,
)


class SimpleStatelessTracingSubscriberRoot:
    def __init__(self, db: SimpleTracingDBSidecar, config: SimpleTracingConfig) -> None:
        self.db = db
        self.config = config

    def notify(self, initializer: Literal[None], /) -> StatelessTracingNotifyStart:
        return SimpleStatelessTracingNotifyStart(
            db=self.db,
            config=self.config,
        )


class SimpleStatelessTracingBroadcasterRoot:
    def receive_notify(
        self, initializer: Literal[None], /
    ) -> StatelessTracingNotifyOnReceived:
        return SimpleStatelessTracingNotifyOnReceived()


if TYPE_CHECKING:
    _: Type[StatelessTracingSubscriberRoot[Literal[None]]] = (
        SimpleStatelessTracingSubscriberRoot
    )
    __: Type[StatelessTracingBroadcasterRoot[Literal[None]]] = (
        SimpleStatelessTracingBroadcasterRoot
    )



import hmac
from typing import TYPE_CHECKING, Optional, Type

from lonelypsp.auth.config import (
    AuthResult,
    ToBroadcasterAuthConfig,
    ToSubscriberAuthConfig,
)
from lonelypsp.auth.set_subscriptions_info import SetSubscriptionsInfo
from lonelypsp.stateful.messages.configure import S2B_Configure
from lonelypsp.stateful.messages.confirm_configure import B2S_ConfirmConfigure
from lonelypsp.stateful.messages.continue_receive import S2B_ContinueReceive
from lonelypsp.stateful.messages.disable_zstd_custom import B2S_DisableZstdCustom
from lonelypsp.stateful.messages.enable_zstd_custom import B2S_EnableZstdCustom
from lonelypsp.stateful.messages.enable_zstd_preset import B2S_EnableZstdPreset
from lonelypsp.stateless.make_strong_etag import StrongEtag


class ToBroadcasterTokenAuth:
    """Allows and produces the authorization header to the broadcaster in
    the consistent form `f"Bearer {token}"`

    In order for this to be secure, the headers must be encrypted, typically via
    HTTPS.
    """

    def __init__(self, /, *, token: str) -> None:
        self.expecting = f"Bearer {token}"
        """The exact authorization header the broadcaster receives"""

    async def setup_to_broadcaster_auth(self) -> None: ...
    async def teardown_to_broadcaster_auth(self) -> None: ...

    def _check_header(self, authorization: Optional[str]) -> AuthResult:
        if authorization is None:
            return AuthResult.UNAUTHORIZED
        if not hmac.compare_digest(authorization, self.expecting):
            return AuthResult.FORBIDDEN
        return self._check_header(authorization)

    async def authorize_subscribe_exact(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
    ) -> Optional[str]:
        return self.expecting

    async def is_subscribe_exact_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return self._check_header(authorization)

    async def authorize_subscribe_glob(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
    ) -> Optional[str]:
        return self.expecting

    async def is_subscribe_glob_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return self._check_header(authorization)

    async def authorize_notify(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        identifier: bytes,
        message_sha512: bytes,
        now: float,
    ) -> Optional[str]:
        return self.expecting

    async def is_notify_allowed(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        identifier: bytes,
        message_sha512: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return self._check_header(authorization)

    async def authorize_stateful_configure(
        self,
        /,
        *,
        tracing: bytes,
        subscriber_nonce: bytes,
        enable_zstd: bool,
        enable_training: bool,
        initial_dict: int,
    ) -> Optional[str]:
        return self.expecting

    async def is_stateful_configure_allowed(
        self, /, *, message: S2B_Configure, now: float
    ) -> AuthResult:
        return self._check_header(message.authorization)

    async def authorize_check_subscriptions(
        self, /, *, tracing: bytes, url: str, now: float
    ) -> Optional[str]:
        return self.expecting

    async def is_check_subscriptions_allowed(
        self, /, *, tracing: bytes, url: str, now: float, authorization: Optional[str]
    ) -> AuthResult:
        return self._check_header(authorization)

    async def authorize_set_subscriptions(
        self, /, *, tracing: bytes, url: str, strong_etag: StrongEtag, now: float
    ) -> Optional[str]:
        return self.expecting

    async def is_set_subscriptions_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        strong_etag: StrongEtag,
        subscriptions: SetSubscriptionsInfo,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return self._check_header(authorization)

    async def authorize_stateful_continue_receive(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        part_id: int,
        url: str,
        now: float,
    ) -> Optional[str]:
        return self.expecting

    async def is_stateful_continue_receive_allowed(
        self, /, *, url: str, message: S2B_ContinueReceive, now: float
    ) -> AuthResult:
        return self._check_header(message.authorization)

    async def authorize_confirm_receive(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        num_subscribers: int,
        url: str,
        now: float,
    ) -> Optional[str]:
        return self.expecting

    async def is_confirm_receive_allowed(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        num_subscribers: int,
        url: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return self._check_header(authorization)

    async def authorize_confirm_missed(
        self, /, *, tracing: bytes, topic: bytes, url: str, now: float
    ) -> Optional[str]:
        return self.expecting

    async def is_confirm_missed_allowed(
        self,
        /,
        *,
        tracing: bytes,
        topic: bytes,
        url: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return self._check_header(authorization)


class ToSubscriberTokenAuth:
    """Allows and produces the authorization header to the subscriber in
    the consistent form `f"Bearer {token}"`

    In order for this to be secure, the clients must verify the header matches
    what they expect and the headers must be encrypted, typically via HTTPS.
    """

    def __init__(self, /, *, token: str) -> None:
        self.expecting = f"Bearer {token}"
        """The authorization header that the subscriber receives"""

    async def setup_to_subscriber_auth(self) -> None: ...
    async def teardown_to_subscriber_auth(self) -> None: ...

    def _check_header(self, authorization: Optional[str]) -> AuthResult:
        if authorization is None:
            return AuthResult.UNAUTHORIZED
        if not hmac.compare_digest(authorization, self.expecting):
            return AuthResult.FORBIDDEN
        return self._check_header(authorization)

    async def authorize_receive(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        identifier: bytes,
        now: float,
    ) -> Optional[str]:
        return self.expecting

    async def is_receive_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        identifier: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return self._check_header(authorization)

    async def authorize_missed(
        self, /, *, tracing: bytes, recovery: str, topic: bytes, now: float
    ) -> Optional[str]:
        return self.expecting

    async def is_missed_allowed(
        self,
        /,
        *,
        tracing: bytes,
        recovery: str,
        topic: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return self._check_header(authorization)

    async def authorize_confirm_subscribe_exact(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
    ) -> Optional[str]:
        return self.expecting

    async def is_confirm_subscribe_exact_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        exact: bytes,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return self._check_header(authorization)

    async def authorize_confirm_subscribe_glob(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
    ) -> Optional[str]:
        return self.expecting

    async def is_confirm_subscribe_glob_allowed(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        recovery: Optional[str],
        glob: str,
        now: float,
        authorization: Optional[str],
    ) -> AuthResult:
        return self._check_header(authorization)

    async def authorize_confirm_notify(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        subscribers: int,
        topic: bytes,
        message_sha512: bytes,
        now: float,
    ) -> Optional[str]:
        return self.expecting

    async def is_confirm_notify_allowed(
        self,
        /,
        *,
        tracing: bytes,
        identifier: bytes,
        subscribers: int,
        topic: bytes,
        message_sha512: bytes,
        authorization: Optional[str],
        now: float,
    ) -> AuthResult:
        return self._check_header(authorization)

    async def authorize_stateful_confirm_configure(
        self, /, *, broadcaster_nonce: bytes, tracing: bytes, now: float
    ) -> Optional[str]:
        return self.expecting

    async def is_stateful_confirm_configure_allowed(
        self, /, *, message: B2S_ConfirmConfigure, now: float
    ) -> AuthResult:
        return self._check_header(message.authorization)

    async def authorize_stateful_enable_zstd_preset(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        compressor_identifier: int,
        compression_level: int,
        min_size: int,
        max_size: int,
        authorization: Optional[str],
        now: float,
    ) -> Optional[str]:
        return self.expecting

    async def is_stateful_enable_zstd_preset_allowed(
        self, /, *, url: str, message: B2S_EnableZstdPreset, now: float
    ) -> AuthResult:
        return self._check_header(message.authorization)

    async def authorize_stateful_enable_zstd_custom(
        self,
        /,
        *,
        tracing: bytes,
        url: str,
        compressor_identifier: int,
        compression_level: int,
        min_size: int,
        max_size: int,
        sha512: bytes,
        now: float,
    ) -> Optional[str]:
        return self.expecting

    async def is_stateful_enable_zstd_custom_allowed(
        self, /, *, url: str, message: B2S_EnableZstdCustom, now: float
    ) -> AuthResult:
        return self._check_header(message.authorization)

    async def authorize_stateful_disable_zstd_custom(
        self, /, *, tracing: bytes, compressor_identifier: int, url: str, now: float
    ) -> Optional[str]:
        return self.expecting

    async def is_stateful_disable_zstd_custom_allowed(
        self, /, *, url: str, message: B2S_DisableZstdCustom, now: float
    ) -> AuthResult:
        return self._check_header(message.authorization)


if TYPE_CHECKING:
    _: Type[ToBroadcasterAuthConfig] = ToBroadcasterTokenAuth
    __: Type[ToSubscriberAuthConfig] = ToSubscriberTokenAuth



import time

from lonelypsp.compat import fast_dataclass


@fast_dataclass
class CristiansStart:
    """The data stored by the initiator"""

    request_sent_at_nano: int


@fast_dataclass
class CristiansResponse:
    """The data sent by the receiver"""

    request_received_at_nano: int
    response_sent_at_nano: int


@fast_dataclass
class CristiansEnd:
    """The data stored by the initiator"""

    request_sent_at_local_nano: int
    request_received_at_remote_nano: int
    response_sent_at_remote_nano: int
    response_received_at_local_nano: int


@fast_dataclass
class CristiansOffset:
    """The offset calculated from the Cristian's algorithm"""

    latency: float
    """estimated one-way latency in milliseconds"""

    offset: float
    """estimated offset of the remote clock compared to the local clock, in milliseconds
    ```
    remote_time - local_time = offset
    local_time = remote_time - offset
    ```
    """


def finish_cristians(
    start: CristiansStart, response: CristiansResponse
) -> CristiansEnd:
    """Finishes the Cristian's algorithm"""
    return CristiansEnd(
        request_sent_at_local_nano=start.request_sent_at_nano,
        request_received_at_remote_nano=response.request_received_at_nano,
        response_sent_at_remote_nano=response.response_sent_at_nano,
        response_received_at_local_nano=time.time_ns(),
    )


def compute_offset(cristians_end: CristiansEnd) -> CristiansOffset:
    """Computes the offset from the Cristian's algorithm"""
    time_processing_nano = (
        cristians_end.response_received_at_local_nano
        - cristians_end.request_sent_at_local_nano
    )
    round_trip_time_nano = (
        cristians_end.response_received_at_local_nano
        - cristians_end.request_sent_at_local_nano
        - time_processing_nano
    )
    latency_nano = round_trip_time_nano / 2
    request_received_at_local_nano = (
        cristians_end.request_sent_at_local_nano + latency_nano
    )
    offset_nano = (
        cristians_end.request_received_at_remote_nano - request_received_at_local_nano
    )
    return CristiansOffset(
        latency=latency_nano / 1_000_000, offset=offset_nano / 1_000_000
    )



from typing import TYPE_CHECKING, Collection, List, Literal, Optional, Type, Union

from lonelypsp.compat import fast_dataclass
from lonelypsp.stateful.constants import (
    PubSubStatefulMessageFlags,
    SubscriberToBroadcasterStatefulMessageType,
)
from lonelypsp.stateful.generic_parser import S2B_MessageParser
from lonelypsp.stateful.parser_helpers import parse_simple_headers
from lonelypsp.stateful.serializer_helpers import (
    MessageSerializer,
    int_to_minimal_unsigned,
    serialize_simple_message,
)
from lonelypsp.sync_io import SyncReadableBytesIO


@fast_dataclass
class S2B_ContinueReceive:
    """
    S2B = Subscriber to Broadcaster
    See the type enum documentation for more information on the fields
    """

    type: Literal[SubscriberToBroadcasterStatefulMessageType.CONTINUE_RECEIVE]
    """discriminator value"""

    identifier: bytes
    """an arbitrary identifier for the notification assigned by the broadcaster; max 64 bytes
    """

    part_id: int
    """which part the subscriber received; acknowledgments must always be in order"""

    authorization: Optional[str]
    """the authorization header that shows the confirmation was sent by the subscriber

    empty strings are converted to None for consistency with http endpoints
    """

    tracing: bytes


_headers: Collection[str] = ("x-identifier", "x-part-id", "authorization", "x-tracing")


class S2B_ContinueReceiveParser:
    """Satisfies S2B_MessageParser[S2B_ContinueReceive]"""

    @classmethod
    def relevant_types(cls) -> List[SubscriberToBroadcasterStatefulMessageType]:
        return [SubscriberToBroadcasterStatefulMessageType.CONTINUE_RECEIVE]

    @classmethod
    def parse(
        cls,
        flags: PubSubStatefulMessageFlags,
        type: SubscriberToBroadcasterStatefulMessageType,
        payload: SyncReadableBytesIO,
    ) -> S2B_ContinueReceive:
        assert type == SubscriberToBroadcasterStatefulMessageType.CONTINUE_RECEIVE

        headers = parse_simple_headers(flags, payload, _headers)
        identifier = headers["x-identifier"]
        if len(identifier) > 64:
            raise ValueError("x-identifier must be at most 64 bytes")

        part_id_bytes = headers["x-part-id"]
        if len(part_id_bytes) > 8:
            raise ValueError("x-part-id must be at most 8 bytes")

        part_id = int.from_bytes(part_id_bytes, "big")

        authorization_bytes = headers["authorization"]
        authorization: Optional[str] = None
        if authorization_bytes != b"":
            try:
                authorization = authorization_bytes.decode("utf-8")
            except UnicodeDecodeError:
                raise ValueError("authorization must be a utf-8 string")

        tracing = headers["x-tracing"]

        return S2B_ContinueReceive(
            type=type,
            identifier=identifier,
            part_id=part_id,
            authorization=authorization,
            tracing=tracing,
        )


if TYPE_CHECKING:
    _: Type[S2B_MessageParser[S2B_ContinueReceive]] = S2B_ContinueReceiveParser


def serialize_s2b_continue_receive(
    msg: S2B_ContinueReceive, /, *, minimal_headers: bool
) -> Union[bytes, bytearray]:
    """Satisfies MessageSerializer[B2S_ContinueNotify]"""
    return serialize_simple_message(
        type=msg.type,
        header_names=_headers,
        header_values=(
            msg.identifier,
            int_to_minimal_unsigned(msg.part_id),
            msg.authorization.encode("utf-8") if msg.authorization is not None else b"",
            msg.tracing,
        ),
        payload=b"",
        minimal_headers=minimal_headers,
    )


if TYPE_CHECKING:
    __: MessageSerializer[S2B_ContinueReceive] = serialize_s2b_continue_receive



from typing import AsyncIterator, Protocol

from lonelypsp.stateless.make_strong_etag import GlobAndRecovery, TopicAndRecovery


class SetSubscriptionsInfo(Protocol):
    """Allows a single, ordered iteration over the topics and urls that a subscriber
    provided in SET_SUBSCRIPTIONS.
    """

    def topics(self) -> AsyncIterator[TopicAndRecovery]:
        """MUST be called before globs() or an error is raised, cannot be called
        multiple times, the corresponding iterable cannot be returned to the start
        """

    def globs(self) -> AsyncIterator[GlobAndRecovery]:
        """MUST be called after topics() or an error is raised, cannot be
        called multiple times, the corresponding iterable cannot be returned to the start
        """



from typing import Generic, Optional, Protocol, TypeVar

from lonelypsp.auth.config import BadAuthResult
from lonelypsp.tracing.shared.tracing_and_followup import TracingAndFollowup

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)


class HandledTrustedNotify(Generic[T], Protocol):
    """Runs on the broadcaster as part of various other sequences, notifying
    subscribers in turn until the final subscriber is reached. Assumes that
    from context the topic that is being notified is known

    T is the next object to use if this ends normally
    """

    def on_unavailable(self) -> None:
        """Called if the subscriber database could not be reached (DBConfig
        returned UNAVAILABLE). Terminates this loop
        """

    def on_exact_subscriber_found(
        self, /, *, url: str
    ) -> "HandledTrustedNotifySendingReceive[HandledTrustedNotify[T]]":
        """Called if the subscriber database found another subscriber that
        is subscribed to this exact topic

        Args:
            url (str): the URL of the exact subscriber
        """

    def on_glob_subscriber_found(
        self, /, *, glob: str, url: str
    ) -> "HandledTrustedNotifySendingReceive[HandledTrustedNotify[T]]":
        """Called if the subscriber database found another subscriber that
        is subscribed to the topic via a glob pattern

        Args:
            glob (str): the glob pattern that the topic matched
            url (str): the URL of the subscriber
        """

    def on_no_more_subscribers(self) -> T:
        """Called if there are no more subscribers to notify, which is the normal
        end to the loop
        """


class HandledTrustedNotifySendingReceive(Generic[T], Protocol):
    """Runs on the broadcaster"""

    def on_sending_receive(
        self, /, *, identifier: bytes
    ) -> TracingAndFollowup["HandledTrustedNotifyReceivedResponse[T]"]:
        """Produces the tracing data to send to the subscriber for processing
        the `RECEIVE`, plus the object on the broadcaster side to track the
        response from the RECEIVE

        Args:
            identifier (bytes): a new identifier for the RECEIVE that was assigned
                by the broadcaster. This will be passed to the subscriber without
                needing to include it in the tracing data. This can be assumed to
                be globally unique

        Returns:
            tracing: the tracing data to send to the subscriber
            followup: the object on the broadcaster side to track the response
        """


class HandledTrustedNotifyReceivedResponse(Generic[T_co], Protocol):
    """Runs on the broadcaster"""

    def on_network_error(self) -> "HandledTrustedNotifyHandleMissedStart[T_co]":
        """Called if no response was received from the subscriber, meaning either
        they could not be reached, didn't understand the protocol used, they
        experienced an error that prevented them from producing a sensible
        response, or the response never reached us due to network-level issues.

        A `MISSED` message will be scheduled if possible and then will move onto
        the next subscriber
        """

    def on_response_received(
        self, /, *, status_code: int
    ) -> "HandleTrustedNotifyResponseAuthResultReady[T_co]":
        """Called if a response was received from the subscriber but before the
        body has been parsed and before the authorization check on that response
        has been performed

        Args:
            status_code (int): the HTTP status code of the response, which isn't
                used for anything but logging
        """


class HandleTrustedNotifyResponseAuthResultReady(Generic[T_co], Protocol):
    def on_bad_receive_response(self) -> "HandledTrustedNotifyHandleMissedStart[T_co]":
        """Called if the response was malformed or otherwise invalid, which
        will schedule a `MISSED` message if possible and then move onto the
        next subscriber
        """

    def on_unsubscribe_immediate_requested(
        self,
    ) -> "HandleTrustedNotifyUnsubscribeImmediateDone[T_co]":
        """Called if the subscriber returned the unsubscribe immediate response,
        which does not require authorization and thus cannot/does not include
        tracing data. The unsubscribe will be processed and then the broadcaster
        will move onto the next subscriber
        """

    def on_bad_receive_auth_result(
        self, /, *, result: BadAuthResult
    ) -> "HandledTrustedNotifyHandleMissedStart[T_co]":
        """Called if the message was correctly formed but the authorization header
        was invalid, which will schedule a `MISSED` message if possible and then
        move onto the next subscriber
        """

    def on_receive_confirmed(self, /, *, tracing: bytes, num_subscribers: int) -> T_co:
        """Called if the response was correctly formed and the auth result succeeded.
        This is done in a single pass, so the tracing data is ready to be deserialized.

        Args:
            tracing (bytes): the tracing data from the subscriber
            num_subscribers (int): the number of subscribers that were notified
                by this RECEIVE, usually 1, but if the subscriber is also itself
                a broadcaster, may be 0 or more.
        """


class HandleTrustedNotifyUnsubscribeImmediateDone(Generic[T_co], Protocol):
    def on_unsubscribe_immediate_success(self) -> T_co:
        """Called when the subscriber has been successfully unsubscribed and
        the broadcaster is ready to move onto the next subscriber
        """

    def on_unsubscribe_immediate_not_found(self) -> T_co:
        """Called if the subscriber had already been removed from the database,
        and the broadcaster is ready to move onto the next subscriber
        """

    def on_unsubscribe_immediate_unavailable(self) -> None:
        """Called when the subscriber could not be unsubscribed because of a
        database issue, which ends the loop
        """


class HandledTrustedNotifyHandleMissedStart(Generic[T_co], Protocol):
    """Runs on the broadcaster"""

    def on_handle_missed_start(self) -> "HandleTrustedNotifyHandleMissedDone[T_co]":
        """Called when about to try scheduling a `MISSED` message before
        proceeding onto something else
        """


class HandleTrustedNotifyHandleMissedDone(Generic[T_co], Protocol):
    """Runs on the broadcaster"""

    def on_handle_missed_skipped(
        self, /, *, recovery: Optional[str], next_retry_at: Optional[float]
    ) -> T_co:
        """Did not schedule a `MISSED` message because either the subscriber did
        not have a recovery url set or the configuration of the broadcaster indicated
        one should not be sent

        Args:
            recovery (str, None): if the subscriber had a place to send MISSED messages,
                that url, otherwise None
            next_retry_at (float, None): if the broadcaster is configured to send missed
                messages, the time at which it was scheduled for, otherwise None
        """

    def on_handle_missed_success(self, recovery: str, next_retry_at: float) -> T_co:
        """Scheduled a `MISSED` message normally

        Args:
            recovery (str): the URL to send the `MISSED` message to
            next_retry_at (float): the time at which the `MISSED` message was scheduled
        """

    def on_handle_missed_unavailable(self, recovery: str, next_retry_at: float) -> T_co:
        """Failed to schedule a `MISSED` message due to an internal issue (the
        database was unavailable). Will still proceed as normal as MISSED messages
        are on a best-effort basis and the core database may still be functioning

        Args:
            recovery (str): the URL to send the `MISSED` message to
            next_retry_at (float): the time at which the `MISSED` message was supposed
                to be scheduled for
        """



import asyncio
from typing import Optional, TypeVar, Union, overload

T = TypeVar("T")
D = TypeVar("D")


@overload
async def cancel_and_check(task: asyncio.Task[T]) -> Optional[T]: ...


@overload
async def cancel_and_check(task: asyncio.Task[T], default: D) -> Union[T, D]: ...


async def cancel_and_check(
    task: asyncio.Task[T], default: Optional[D] = None
) -> Optional[Union[T, D]]:
    """If the task is finished, returns the result. If the task is errored,
    raises the error. If the task is cancelled, returns the default value.
    If the task is running, schedules asyncio.CancelledError to be raised
    at the current checkpoint and waits for the callbacks to finish, then
    rechecks the state of the task (which will no longer be running). Takes
    an arbitrary number of event loop iterations to complete.

    When using this, _if_ this is not canceled, then this always
    retrieves exceptions

    This is surprisingly difficult to do correctly
    """
    if not task.cancel():
        return task.result()

    try:
        await task
    except BaseException:
        ...

    if task.cancelled():
        return default
    return task.result()
